import dataclasses
import logging
import os
import pathlib
import tomllib
from typing import Dict, List, Union

import tomli_w

LOGGER = logging.getLogger(__name__)

import rgpio

@dataclasses.dataclass(frozen=True)
class Pin:
	chip: str
	number: int
	name: str
	@property
	def id(self) -> str:
		return f"{self.chip}-{self.number}"

@dataclasses.dataclass(frozen=True)
class Relay:
	pin: Pin
	name: str
	state: bool

	@property
	def id(self) -> str:
		return f"{self.pin.id}"

def _deserialize(data: List[Dict[str, str]]) -> List[Relay]:
	"Deserialize a list of relays from the config."
	return [Relay(
		name=relay["name"],
		pin=Pin(
			chip=relay["chip"],
			number=int(relay["pinnumber"]),
			name=relay["pinname"],
		),
		state=False) for relay in data]

def _serialize(data: List[Relay]) -> List[Dict[str, str]]:
	"Serialize a list of relays to the config."
	LOGGER.info("Serializing %s", data)
	return [{
		"chip": relay.pin.chip,
		"name": relay.name,
		"pinnumber": str(relay.pin.number),
		"pinname": relay.pin.name,
	} for relay in data]

class Manager:
	"Class for interacting with relays."
	def __init__(self, configpath: pathlib.Path, has_fakes: bool):
		try:
			self.config = _read_config(configpath)
		except FileNotFoundError:
			self.config = {
				"rgpio": {"host": "localhost", "port": 8889},
				"relays": [{
					"chip": -1,
					"pin": "FAKE-P1",
					"name": "Fake",
				}],
			}
			# Immediately write out the default config
			_write_config(configpath, self.config)
		self.configpath = configpath
		self.has_fakes = has_fakes

		# Connection to single-board computer GPIO
		self.sbc = None

	def __iter__(self):
		"Provide an iterator for the relays."
		return iter(self.relays)

	def connect(self) -> None:
		self.sbc = rgpio.sbc(
			host=self.config["rgpio"]["host"],
			port=self.config["rgpio"]["port"],
		)

	def chips_list(self) -> List[str]:
		if self.has_fakes:
			return ["fakeA", "fakeB"]

	def get_relay_by_pin_id(self, pin_id: str) -> Relay:
		for relay in self.relays:
			if relay.pin.id == pin_id:
				return relay
		return None

	def pins_list(self) -> List[Pin]:
		if self.has_fakes:
			return [
				Pin(chip="fakeA", number=1, name="CON2-P1"),
				Pin(chip="fakeA", number=2, name="CON2-P2"),
				Pin(chip="fakeA", number=3, name="CON2-P10"),
				Pin(chip="fakeB", number=1, name="CON2-P3"),
				Pin(chip="fakeB", number=2, name="CON2-P7"),
			]


	def relay_add(self, chip: str, pinnumber: int, name: str) -> None:
		LOGGER.info("Creating relay %s %s with name %s", chip, pinnumber, name)
		self.relays.append(Relay(
			name=name,
			pin=Pin(
				chip=chip,
				number=pinnumber,
				name=name,
			),
			state=False))
		self._write_config()

	def relay_set(self, relay: Relay, state: bool) -> None:
		if not relay.pin.chip.startswith("fake"):
			pass

	def shutdown(self) -> None:
		_write_config(self.configpath, self.config)


def _read_config(configpath: pathlib.Path) -> None:
	with open(configpath, "rb") as f:
		content = tomllib.load(f)
		LOGGER.info("Read config from %s", configpath)
		return content

def _write_config(configpath: pathlib.Path, config) -> None:
	with open(configpath, "wb") as f:
		tomli_w.dump(config, f)
	LOGGER.info("Wrote config file %s", configpath)

